import sys

def validate_electrode_list(electrode_list):
    """Validate electrode configuration exists if any electrode value is not one of -1 , 0  or 1
       Args:
        electrode_list: electrode configuration list 
       Returns:
        None 
    """
    for elec in electrode_list:
        if elec not in [-1, 0, 1]:
            print("Invalid electrode configuration. Elements must be either -1, 0, or 1.")
            sys.exit(4)




import numpy as np
import sys
from metastim import lead_selector


class MetaStimUtil:

    # FUNCTION = field_sd(phi (req))
    # Short description: this is an auxiliary function that calculates SD from phi
    # req = required
    # SD = second (discrete) difference operation
    # INPUTS:
    # phi, electric potentials [in V] in a # points x # axon array by format
    # OUTPUT:
    # SD, 11 second differences of electric potentials per axon [in mV]
    # (format: 11 x # axons)
    # NOTES:
    # - each axon in this demo has the same number of nodes / points
    # - the code should be generalized so that each axon can have different # pts / axon
    # - however, by this part in the code, only 11SD are calculated per axon, so data can be held in a 2D array    
    @classmethod
    def get_field_sd(cls, num_axons, phi_axon):
        sd_axon = 1e3 * np.diff(phi_axon, n=2, axis=0)  # V => mV
        sd_max_indx = np.argmax(sd_axon, axis=0)  # row index per column where SD is max

        window_size = 11  # window size
        nn = int((window_size - 1) / 2)  # number of neighbors to left/right of max
        sd_11_axon = np.zeros((window_size, sd_axon.shape[1]))  # pre-allocate 11SD for each column

        for k in range(0, num_axons):
            # define the bounds of the window
            w_indx_l = np.maximum(sd_max_indx[k] - nn, 0)  # first index of window
            w_indx_r = np.minimum(sd_max_indx[k] + nn, sd_axon.shape[0] - 1)  # last index of window

            # calculate padding
            pad_l = nn - (sd_max_indx[k] - w_indx_l)
            pad_r = nn - (w_indx_r - sd_max_indx[k])

            # pad the values within the window
            sd_11_axon[:, k] = np.pad(sd_axon[w_indx_l:w_indx_r + 1, k], (pad_l, pad_r), 'constant')

        return sd_11_axon

    # FUNCTION = field_shape(SD (req))
    # Short description: this is an auxiliary function that assigns a shape classification of potentials for each axon
    # req = required
    # INPUT:
    # SD, 11 second differences of electric potentials per axon [in mV]
    # (format: 11 x # axons)
    # OUTPUT:
    # Fields Shape (FS), shape classification for 11SD for each axon (1 x # axons)
    # FS can be 1, 2, or 3
    # NOTES:
    # -this is a stopgap based on a simple thresholding of a feature
    # -a more advanced classification will be implemented in a future code ver
    @classmethod
    def get_field_shape(cls, num_axons, sd_11_axon):
        fs_axon = np.zeros((num_axons,))

        # calculate maximum SD for each axon
        # calculate minimum SD for each axon
        # calculate absolute value of ratio of max SD / min SD
        sd_max = np.max(sd_11_axon, axis=0)
        sd_min = np.min(sd_11_axon, axis=0)
        sd_rat = np.abs(np.divide(sd_max, sd_min))
        fs_axon[sd_rat >= 2.55] = 1
        fs_axon[(sd_rat >= 1.15) & (sd_rat <2.55)] = 2
        fs_axon[sd_rat < 1.15] = 3
        return fs_axon


    # FUNCTION = axon2lead_dist(lead (rew), axonCoord (req))
    # Short description: this is an auxiliary function that calculates the distance of each axon to the lead
    # req = required
    # INPUT:
    # lead, lead geometry
    # -for this demo, we only need the lead's radius
    # axon_coord, coordinates of axon 
    # OUTPUT:
    # d, distance from lead to axon (# axons x 1)
    # NOTES:
    # -this calculation works for an ideal setup where the lead and axons are perfectly parallel to each other and aligned with the z axis
    # -future version of this code will calculate d for arbitrary lead angles and non-straight axon trajectories
    @classmethod
    def get_axon_to_lead_dist(cls, lead_radius, x_axon, y_axon):
        # calculate distance of axon to z axis at xy = (0,0)
        axon_radius = np.sqrt(np.min(x_axon, axis=0) ** 2 + np.min(y_axon, axis=0) ** 2)
        axon_distance = axon_radius - lead_radius
        axon_distance[axon_distance < 0] = 'NaN'
        return axon_distance


    @classmethod
    def get_lead_radius(cls, lead_id, electrode_list):
        leadselector =  lead_selector.LeadSelector('DBSLead-smry.csv')
        leads = leadselector.load_leads();        
        
        if lead_id not in leads.keys():
            print(f"Invalid lead specified. Lead Id must be  of {leads.keys()}.")
            sys.exit(5)
        else:
            lead = leads.get(lead_id)
            if lead.no != len(electrode_list) :
                print(f"Invalid electrode configuration. {lead_id} contains {lead.no} electrods")
                sys.exit(6)

            # get radius 
            return lead.re


import numpy as np
import sys
from importlib import resources
from keras.models import model_from_json
from joblib import dump, load

from metastim import field_ann 
from metastim.utils import MetaStimUtil
from metastim import visualization as vis



class AxonANN:
    """calculates the voltage required to activate the axons of neurons"""

    def __init__(self, electrode_list, pulse_width , stimulation_amp, num_axons=10, axon_diameter=6):
        self.electrode_list = electrode_list
        self._validate_electrode_list(electrode_list)        
        self.pulse_width = pulse_width
        self._validate_pulse_width(pulse_width)        
        self.stimulation_amp = stimulation_amp
        self._validate_stimulation_amp(stimulation_amp)        
        # TODO: validation for num_axons
        self.num_axons = num_axons
        self._validate_num_axons(num_axons)
        self.axon_diameter = axon_diameter
        self._validate_axon_diameter(axon_diameter)
    
    def axon_ann(self, x_axon, y_axon, z_axon, lead_radius, threshold = None):
        """Predict axon activation based on electric potentials
           Output: (axon_activation, axon_thresholds) 
        """                
        field_ann_model = field_ann.FieldANN(self.electrode_list)
        phi_axon = field_ann_model.field_ann(x_axon, y_axon, z_axon)
        
        # load ann model        
        with resources.open_text("metastim.axon-ann-model", "ann-axon-settings.json") as settings_file:
            json_data = settings_file.read()
            axon_model = model_from_json(json_data)

        #load weights
        with resources.open_text("metastim.axon-ann-model", "ann-axon-weights.h5") as weights_file:
            axon_model.load_weights(weights_file.name)

        # load standard scalar for inputs
        with resources.open_text("metastim.axon-ann-model", "ann-axon-input-std.bin") as std_file:
            sc_axon = load(std_file.name)

        # sd_11_axon
        sd_11_axon = MetaStimUtil.get_field_sd(self.num_axons, phi_axon)

        # fx_axon
        fs_axon = MetaStimUtil.get_field_shape(self.num_axons, sd_11_axon)

        # axon_distance
        axon_distance = MetaStimUtil.get_axon_to_lead_dist(lead_radius, x_axon, y_axon)

        self._validate_axon_distance(axon_distance)
        
        # organize inputs to Axon ANN
        o = np.ones((self.num_axons,))
        x_axon_ann_raw = np.column_stack((fs_axon, o * self.axon_diameter, o * self.pulse_width, axon_distance, np.transpose(sd_11_axon)))

        # standardize inputs for Axon ANN
        x_axon_ann = sc_axon.transform(x_axon_ann_raw)

        # evaluate the Axon ANN model
        y_axon_ann = np.exp(axon_model.predict(x_axon_ann).reshape(-1))        
        axon_activation = (y_axon_ann <= self.stimulation_amp).astype(int)        

        if threshold == None:
            threshold = False

        if threshold:
            axon_thresholds = np.round(y_axon_ann, 2)
            return axon_activation, axon_thresholds
        
        return axon_activation, 
        

    def __repr__(self):
        properties = ", ".join(f"{key}='{value}'" for key, value in vars(self).items())
        return f"{type(self).__name__}({properties})"
    
    def __str__(self):        
        return self.__repr__()
    
        
    def _validate_axon_diameter(self, axon_diameter):
        """Axon Diameter(D) validate:
            if axon_diameter <= 0  print error and exit 
            if  1.5 > axon_diameter < 15  print Warning and continue
        """
        if axon_diameter <= 0:
            print('Error: Negative or Zero fiber diameter (D), D must be non-zero and positive (> 0).')
            sys.exit(1)        
        if axon_diameter < 1.5 or axon_diameter > 15:
            print('Warning! Accuracy may be degraded for fiber diameters outside of 1.5-15um.')              

    def _validate_pulse_width(self, pulse_width):        
        """stimulus pulse width(pw)  pulse_width validation:
            if pw <= 0  print Error and exit
            if  30 < pw > 500 : issue Warning continue
        """ 
        if pulse_width <= 0:
            print('Error: Negative or Zero pulse width (PW)! PW must be non-zero and positive (> 0).')
            sys.exit(2)        
        if pulse_width < 30 or pulse_width > 500:
            print('Warning! Accuracy may be degraded for pulse widths outside of 30-500us.')            
        
    def _validate_electrode_list(self, electrode_list):
        for elec in electrode_list:
            if elec not in [-1, 0, 1]:
                print("Invalid electrode configuration. Elements must be either -1, 0, or 1.")
                sys.exit(4)
    
    def _validate_lead(self, lead_id):
        if lead_id not in self.leads.keys():
            print(f"Invalid lead specified. Lead Id must be  of {self.leads.keys()}.")
            sys.exit(5)
        else:
            lead = self.leads.get(lead_id)
            if lead.no != len(self.electrode_list) :
                print(f"Invalid electrode configuration. {lead_id} contains {lead.no} electrods")
                sys.exit(6)

            # get radius 
            self.lead_radius = lead.re

    def _validate_num_axons(self, num_axons):
        pass
        
    def _validate_stimulation_amp(self, stimulation_amp):
        """use magnitude of stimulation_amp
        """
        self.stimulation_amp = abs(stimulation_amp)

    def _validate_axon_distance(self, axon_distance):
        if (axon_distance.any() < 0.5  or axon_distance.any() > 9):
            print("Warning! Accuracy may be degraded as the minimum distance between axon and lead is out of range (0.5mm - 9mm)")
    

def main():
    """This exists for testing this module"""
    lead_id = '6172'
    electrode_list = [1, 0, 0, 0, -1, 0, 0, 0]
    stimulation_amp = 3
    pulse_width = 90
    num_axons = 10
    min_distance = 1
    max_distance = 5
    axon_diameter = 6

    lead_radius = MetaStimUtil.get_lead_radius(lead_id, electrode_list)

    inl = 100 * axon_diameter / 1e3 # distance between nodes on an axon

    z_base = np.arange(-5, 16, inl)
    num_axon_nodes = z_base.shape[0]

    x_axon = np.repeat(np.linspace(min_distance, max_distance, num=num_axons), num_axon_nodes).reshape(num_axon_nodes, num_axons, order='F') + lead_radius
    y_axon = np.zeros(x_axon.shape)
    z_axon = np.repeat(z_base, num_axons).reshape(num_axon_nodes, num_axons)

    field_ann_model = field_ann.FieldANN(electrode_list)    
    axon_ann_model = AxonANN(electrode_list, pulse_width, stimulation_amp, num_axons, axon_diameter)
    
    phi_axon = field_ann_model.field_ann(x_axon, y_axon, z_axon)
    axon_act = axon_ann_model.axon_ann(x_axon, y_axon, z_axon, lead_radius)

    visualization = vis.Visualization(lead_id, stimulation_amp, num_axons, x_axon, z_axon, phi_axon, axon_act)
    visualization.visualize1(electrode_list)

if __name__ == "__main__":
    main()


from importlib import resources
from joblib import load
from keras.models import model_from_json
from sklearn.preprocessing import StandardScaler
from metastim import validations
import numpy as np
import matplotlib.pyplot as plt
import sys


class FieldANN:
    def __init__(self, electrode_config):
        if len(electrode_config) == 0:
            print("Invalid electrode configuration. Elements configuration is empty.")
            sys.exit(4)
        else:                
            validations.validate_electrode_list(electrode_config)
            self._electrode_config = electrode_config
            self._model = None
            self._std_scaler = None
            self.__load_model()

        
    def __load_model(self):
        """loads field ANN files and creates model"""
        num_elec_on = np.sum(np.abs(self._electrode_config))  # Total number of electrodes on
        model_file =  f'ann-field-ec{num_elec_on}-settings.json'  
        weight_file = f'ann-field-ec{num_elec_on}-weights.h5'
        std_sca_file = f'ann-field-ec{num_elec_on}-input-std.bin'
        
        with resources.open_text("metastim.field-ann-models", model_file) as f:
            model_json = f.read()
            self._model = model_from_json(model_json)
        
        with resources.open_binary("metastim.field-ann-models", weight_file) as wf:
            self._model.load_weights(wf.name)              

        with resources.open_binary("metastim.field-ann-models", std_sca_file) as ssf:            
            self._std_scaler = load(ssf.name)

    def predict_field(self, x, y, z):
        """evluate the model 
           Args:
            x: x vector  
            y: y vector
            z: z vector 
           Returns:
            y_model 
        """         
        xyz = np.column_stack((x, y, z))
        num_points = z.shape[0]
        x_model_raw = np.column_stack((np.tile(self._electrode_config, (num_points, 1)), xyz))
        x_model = self._std_scaler.transform(x_model_raw)
        y_model = np.exp(self._model.predict(x_model).reshape(-1)) - 1
        return y_model

    @property
    def electrode_config(self):
        return self._electrode_config
    
    @electrode_config.setter
    def electrode_config(self, value):
        self._electrode_config = value
        self.__load_model()
    
    def visualize_field(self, x, y, z, stim_amp):
        """visualize field using matplotlib
           Args:              
            x: x vector  
            y: y vector
            z: z vector
            stim_amp: stimulation amplitude            
           Returns:
            None
        """
        font = {'family': 'serif', 'color': 'black', 'size': 20}
        plt.plot(z, stim_amp * self.predict_field(x, y, z), 'k-', linewidth=1)
        plt.title('Sample field calculation', fontdict=font)
        plt.xlabel('z (mm)', fontdict=font)
        plt.ylabel('$\Phi$ (V)', fontdict=font)
        plt.show()


    def field_ann(self, x_axon, y_axon, z_axon, num_axons=10):
        """This is a main function that calculates the electric potentials across axons
           Args:
           electrode_list: electrode configuration(s) (1, # electrodes) 0 is off, 1 is on and positive, -1 is on and negative
           axon_coord: xyz coordinates of each axon (3 x # points per axon x # axons)
           amp, stimulation amplitude in Volts
           lead, lead model (optional, Model 6172 is the selectable option right now)
           OUTPUT:
             phi, electric potentials from Field ANN for each axon [in V]
           NOTES:
                each axon in this demo has the same number of nodes / points
                the code should be generalized so that each axon can have different # pts / axon
                this could be done with a struct or a n x 4 matrix [axon ID, x, y, z], where n = # of points across all axons 
        """
        ec = np.array(self._electrode_config) # electrode configuration (+1, -1, or 0)
        num_electrodes = ec.shape[0] # total number of electrodes
        num_electrodes_on = np.sum(np.abs(ec))
        # x_axon, y_axon, z_axon = self.axon_coord()
        # directories and filenames

        # ----- Load Field ANN files ---
        field_ann_setting_file = f'ann-field-ec{num_electrodes_on}-settings.json'
        field_ann_weight_file = f'ann-field-ec{num_electrodes_on}-weights.h5'
        field_ann_std_file = f'ann-field-ec{num_electrodes_on}-input-std.bin'

        # ----- LOAD MODEL -----
        # load ann model
        
        with resources.open_text("metastim.field-ann-models", field_ann_setting_file) as settings_file:
            json_data = settings_file.read()
            field_model = model_from_json(json_data)

        #load weights
        with resources.open_binary("metastim.field-ann-models", field_ann_weight_file) as weight_file:
            field_model.load_weights(weight_file.name)

        # load standard scalar for inputs
        with resources.open_binary("metastim.field-ann-models", field_ann_std_file) as std_file:            
            sc_field = load(std_file.name)

        # Calculate Potentials from Field ANN
        phi_axon = np.zeros(x_axon.shape)

        for k in range(0, num_axons):
            # organize inputs
            num_nodes = x_axon[:,k].shape[0]
            xyz_axon = np.column_stack((x_axon[:,k], y_axon[:,k], z_axon[:,k]))
            x_field_raw = np.column_stack((np.tile(ec, (num_nodes,1)), xyz_axon)) 

            # standardize inputs
            x_field = sc_field.transform(x_field_raw)

            # evaluate the model
            y_field = np.exp(field_model.predict(x_field).reshape(-1)) - 1 
            phi_axon[:,k] = y_field

        return phi_axon
        


def main():
    electrode_config = np.array([0, 1, 1, 1, 1, 1, 1, 0])  # Electrode configuration (+1, -1, or 0)
    stim_amp = 3  # Stimulation amplitude in Volts

    # Specify x, y and z values for field calculation
    z = np.linspace(-5, 16, num=100)
    x = 1 * np.ones(z.shape)
    y = 1 * np.ones(z.shape)



    # Create an instance of FieldANN class
    field_calculator = FieldANN(electrode_config)

    # y_model = field_calculator.predict_field(x, y, z)
    
    # Visualize the field calculation

    field_calculator.visualize_field(x, y, z, stim_amp)

    electrode_config = np.array([1, 0, 0, 0, -1, 0, 0, 0])
    field_calculator.electrode_config = electrode_config
    field_calculator.visualize_field(x, y, z, stim_amp)

    field_calculator.visualize_field(x, y, z, 5)



if __name__ == "__main__":
    main()


import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.image as image
import os
import sys

from metastim import lead_selector as ls


class Visualization:
    def __init__(self, lead_id, stimulation_amp, num_axons, x_axon, z_axon, phi_axon, axon_activation):
        lead_selector =  ls.LeadSelector('DBSLead-smry.csv')
        self.leads = lead_selector.load_leads();          
        # self.lead_radius = lead_radius
        self._validate_lead(lead_id)
        self.stimulation_amp = stimulation_amp
        self.num_axons = num_axons
        self.x_axon = x_axon
        self.z_axon = z_axon
        self.phi_axon = phi_axon
        self.axon_activation = axon_activation

    def visualize1(self):
            font = {'family':'serif', 'color':'black', 'size':20}
            
            f, (ax1, ax2) = plt.subplots(1, 2)
            h_lead = self.z_axon[-1,0] - self.z_axon[0,0]
            ax1.add_patch(patches.Rectangle((-self.lead_radius, self.z_axon[0,0]), 2*self.lead_radius, h_lead, linewidth=1, edgecolor='k', facecolor='k'))
            ax1.set_xlim([-1,10])
            ax1.set_ylim([self.z_axon[0,0], self.z_axon[-1,0]])
            for k in range(0, self.num_axons):
                if self.axon_activation[k] > 0:
                    ax1.plot([self.x_axon[0,k], self.x_axon[0,k]], [self.z_axon[0,0], self.z_axon[-1,0]], 'g-', linewidth=1) # blue is active
                else:
                    ax1.plot([self.x_axon[0,k], self.x_axon[0,k]], [self.z_axon[0,0], self.z_axon[-1,0]], 'k-', linewidth=1, alpha=0.25) # black is inactive
                
            ax1.set_title('axons & lead', fontdict = font)
            ax1.set_xlabel('node', fontdict = font)
            ax1.set_ylabel('$\Phi$ (V)', fontdict = font)

            for k in range(0, self.num_axons):
                if self.axon_activation[k] > 0:
                    ax2.plot(self.stimulation_amp * self.phi_axon[:,k], 'g-', linewidth=1) # blue is active
                else:
                    ax2.plot(self.stimulation_amp * self.phi_axon[:,k], 'k-', linewidth=1, alpha=0.25) # black is inactive
                
            ax2.set_title('potentials across axons', fontdict = font)
            ax2.set_xlabel('node', fontdict = font)
            ax2.set_ylabel('$\Phi$ (V)', fontdict = font)
            
            plt.show()

    def _validate_lead(self, lead_id):
        if lead_id not in self.leads.keys():
            print(f"Invalid lead specified. Lead Id must be  of {self.leads.keys()}.")
            sys.exit(7)
        else:
            lead = self.leads.get(lead_id)         
            # get radius 
            self.lead_radius = lead.re

    def visualize2(self, ec):
            font = {'family':'serif', 'color':'black', 'size':20}
            
            f, (ax1, ax2) = plt.subplots(1, 2)
            
            ax1.axis([0, 25, 0, 25])

            # draw controll panel

            x_offset = 5
            y_offset = 5

            cell_labels = ['1', '2', '3', '4', '5', '6', '7', '8']
            x_positions = [1, 1, 3, 5, 1, 3, 5, 1] 
            y_positions = [1, 4, 4, 4, 7, 7, 7, 10]

            for i in range(8):
                color = None
                if ec[i] == 1:
                    color = 'red'
                elif ec[i] == -1:
                    color = 'skyblue'
                elif ec[i] == 0:
                    color = 'gray'

                x = x_positions[i] + x_offset
                y = y_positions[i] + y_offset

                if ( i == 0 or i == 7):        
                    # Row 1 & 4
                    shape = patches.Rectangle((x , y ), width=5.8, height=2, color=color)
                    ax1.add_patch(shape)
                    if i == 0:
                        ax1.text(x + 3, y + 1, cell_labels[i], ha='center', va='center', fontsize=12)
                    if i == 7:
                        ax1.text(x + 3, y + 1, cell_labels[i], ha='center', va='center', fontsize=12)
                else:
                    # Row 2 , 3, 4, 5, 6, 7
                    shape = patches.Rectangle((x, y), width=1.6, height=2, color=color, linewidth=2)
                    ax1.add_patch(shape)
                    ax1.text(x + 1, y + 1, cell_labels[i], ha='center', va='center', fontsize=12)

            # draw lead image 

            # lead_image = image.imread('./images/lead2.png')
            # ax1.imshow(lead_image)

            # leads             
            for k in range(0, self.num_axons):
                if self.axon_activation[k] > 0:
                    ax1.plot([self.x_axon[0,k] + 16, self.x_axon[0,k] + 16], [self.z_axon[0,0] + 5, self.z_axon[-1,0]], 'g-', linewidth=1) # blue is active
                else:
                    ax1.plot([self.x_axon[0,k] + 16, self.x_axon[0,k] + 16], [self.z_axon[0,0] + 5, self.z_axon[-1,0]], 'k-', linewidth=1, alpha=0.25) # black is inactive
                
            ax1.set_title('axons & lead', fontdict = font)
            ax1.set_xlabel('node', fontdict = font)
            ax1.set_ylabel('$\Phi$ (V)', fontdict = font)

            # Pottentials

            for k in range(0, self.num_axons):
                if self.axon_activation[k] > 0:
                    ax2.plot(self.stimulation_amp * self.phi_axon[:,k], 'g-', linewidth=1) # blue is active
                else:
                    ax2.plot(self.stimulation_amp * self.phi_axon[:,k], 'k-', linewidth=1, alpha=0.25) # black is inactive
                
            ax2.set_title('potentials across axons', fontdict = font)
            ax2.set_xlabel('node', fontdict = font)
            ax2.set_ylabel('$\Phi$ (V)', fontdict = font)
            
            plt.show()


    def visualize(self, ec):
            font = {'family':'serif', 'color':'black', 'size':20}
            
            f, (ax1, ax2, ax3) = plt.subplots(1, 3)
            
            ax1.axis([0, 25, 0, 25])

            # draw controll panel

            x_offset = 5
            y_offset = 5

            cell_labels = ['1', '2', '3', '4', '5', '6', '7', '8']
            x_positions = [1, 1, 3, 5, 1, 3, 5, 1] 
            y_positions = [1, 4, 4, 4, 7, 7, 7, 10]

            for i in range(8):
                color = None
                if ec[i] == 1:
                    color = 'red'
                elif ec[i] == -1:
                    color = 'skyblue'
                elif ec[i] == 0:
                    color = 'gray'

                x = x_positions[i] + x_offset
                y = y_positions[i] + y_offset

                if ( i == 0 or i == 7):        
                    # Row 1 & 4
                    shape = patches.Rectangle((x , y ), width=5.8, height=2, color=color)
                    ax1.add_patch(shape)
                    if i == 0:
                        ax1.text(x + 3, y + 1, cell_labels[i], ha='center', va='center', fontsize=12)
                    if i == 7:
                        ax1.text(x + 3, y + 1, cell_labels[i], ha='center', va='center', fontsize=12)
                else:
                    # Row 2 , 3, 4, 5, 6, 7
                    shape = patches.Rectangle((x, y), width=1.6, height=2, color=color, linewidth=2)
                    ax1.add_patch(shape)
                    ax1.text(x + 1, y + 1, cell_labels[i], ha='center', va='center', fontsize=12)
            ax1.set_title('Control Panel', fontdict = font)

            # draw lead image 
            
            ax2.set_xlim(0, 2)
            ax2.set_ylim(0, 10)
            ax2.set_aspect('equal')


            main_body_width = 0.3
            main_body_height = 6
            main_body = patches.Rectangle((0.85, 2), main_body_width, main_body_height, linewidth=1, edgecolor='gray', facecolor='white')
            ax2.add_patch(main_body)


            rect_height = 1
            spacing = 0.5
            gray_rects = [
                (0.85, 6.5),  # First rectangle at the top
                (0.85, 5),    # Second rectangle
                (0.85, 3.5),  # Third rectangle
                (0.85, 2)     # Fourth rectangle
            ]

            for rect in gray_rects:
                if rect[1] == 5 or rect[1] == 3.5:
                    cylinder_width = main_body_width / 3
                    ax2.add_patch(patches.Rectangle((rect[0], rect[1]), cylinder_width, rect_height, linewidth=1, edgecolor='gray', facecolor='gray'))
                    ax2.add_patch(patches.Rectangle((rect[0] + 2 * cylinder_width, rect[1]), cylinder_width, rect_height, linewidth=1, edgecolor='gray', facecolor='gray'))
                elif rect[1] == 2:
                    ax2.add_patch(patches.Rectangle((rect[0], rect[1]), main_body_width, rect_height / 1.5, linewidth=1, edgecolor='gray', facecolor='gray'))
                else:
                    ax2.add_patch(patches.Rectangle((rect[0], rect[1]), main_body_width, rect_height , linewidth=1, edgecolor='gray', facecolor='gray'))

            
            ellipse_width = main_body_width
            ellipse_height = 0.35
            ellipse = patches.Ellipse((1.0052, 2), ellipse_width, ellipse_height, angle=0, linewidth=1, edgecolor='gray', facecolor='white')
            ax2.add_patch(ellipse)

            
            bottom_black = patches.Ellipse((1.0052, 2), ellipse_width, ellipse_height, angle=0, linewidth=1, edgecolor='gray', facecolor='gray')
            ax2.add_patch(bottom_black)
            ax2.set_title('lead', fontdict = font)
            
            
            # Plot Pottentials

            for k in range(0, self.num_axons):
                if self.axon_activation[k] > 0:
                    ax3.plot(self.stimulation_amp * self.phi_axon[:,k], 'g-', linewidth=1) # blue is active
                else:
                    ax3.plot(self.stimulation_amp * self.phi_axon[:,k], 'k-', linewidth=1, alpha=0.25) # black is inactive
                
            ax3.set_title('potentials across axons', fontdict = font)
            ax3.set_xlabel('node', fontdict = font)
            ax3.set_ylabel('$\Phi$ (V)', fontdict = font)
            
            plt.show()



import csv
from importlib import resources

from metastim import lead

class LeadSelector():
    def __init__(self, csv_file):
        self.csv_file = csv_file
        self.leads = self.load_leads()

    def load_leads(self):
        leads = {}
        with resources.open_text("metastim", self.csv_file) as csv_resource:
            with open(csv_resource.name) as csv_file:
                reader = csv.DictReader(csv_file)
                for row in reader:
                    lead_id = row['Lead']
                    lead_type = row['Type']
                    company = row['Company']
                    no = int(row['no'])
                    he = float(row['h_e [mm]'])
                    re = float(row['r_e [mm]'])
                    ae = float(row['a_e [deg]']) if row['a_e [deg]'] != 'N/A' else None
                    ies = float(row['ies [mm]'])
                    htip = float(row['h_tip [mm]'])
                    leads[lead_id] = lead.Lead(lead_id, lead_type, company, no, he, re, ae, ies, htip)
        return leads   

    def select_lead(self, id):
        return self.leads.get(id)




# class Represents Lead

class Lead:
    # initializr
    def __init__(self, lead_id = "6172", type = "Directional", company = "Abbot Laboratories", no = 8, he = 1.5, re = 0.635, ae = 104, ies = 0.5, htip = 1.5):
        self.lead_id = lead_id
        self.type = type
        self.company = company
        self.no = no
        self.he = he
        self.re = re
        self.ae = ae
        self.ies = ies
        self.htip = htip
    
    def __str__(self):
        return f"Lead: {self.lead_id}, Lead Type: {self.type}, Company: {self.company}, No: {self.no}, he : {self.he}, re : {self.re}, ae : {self.ae}, ies: {self.ies}, htip : {self.htip}"

    def __repr__(self):
        return f"Lead(lead_id='{self.lead_id}', type='{self.type}', company='{self.company}', no={self.no}, he={self.he}, re={self.re}, ae={self.ae}, ies={self.ies}, htip={self.htip})"